"""
CPM stands for Color Palette Algorithms
https://github.com/gejirz/PixPalette-algorithms - you can visit this script repository on github
In order for you to use functions from here, I tried my best to briefly document my work.
Feel free to contact me for any help! mail - superavb222@gmail.com
"""
from sys import path
from pathlib import Path
path.append(str(Path(__file__).parent.resolve()))
from skimage.io import imread
from skimage.transform import resize
from skimage.util import img_as_ubyte, invert
from numpy import full, array, ndarray, asarray, zeros, amin, amax, clip
from PIL import Image, ImageDraw, ImageFont
from PIL.ImageOps import scale
from typing import List

from MMCQ import palette
from ImageCreation import image

__version__ = "0.0.1"


def _DEV_formatting(img: ndarray, pix_limit: int, img_w: int, img_h: int, img_c: int, img_d: str) -> ndarray:
    """
    Formatting algorithm for a image, will resize it to fit the pix limit, delete/convert data
    Input:
    :param img: input image (numpy array) to reformat
    :param pix_limit: limit of pixels for the reformatted image
    :param img_w: width of the image
    :param img_h: height of the image
    :param img_c: amount of channels of the image
    :param img_d: dtype of the image
    Output:
    A image in numpy array form
    """

    img_d = img.dtype
    if img_w * img_h >= pix_limit:
        m = (pix_limit / (img_w * img_h)) ** 0.5
        img_h = int(img.shape[0] * m)
        img_w = int(img.shape[1] * m)
        img = resize(img, (img_h, img_w), preserve_range=True)
        img = asarray(img, dtype=img_d)
        if img_d == 'uint8' and img_c == 3:
            return img

    if img_c == 4:
        new_img = zeros((img_h, img_w, 3), dtype=img_d)
        new_img[:, :, 0] = img[:, :, 0]
        new_img[:, :, 1] = img[:, :, 1]
        new_img[:, :, 2] = img[:, :, 2]
        img = new_img
    elif img_c == 2:
        new_img = zeros((img_h, img_w, 3), dtype=img_d)
        new_img[:, :, 0] = img[:, :, 0]
        new_img[:, :, 1] = img[:, :, 1]
        new_img[:, :, 2] = img[:, :, 0]
        img = new_img

    if img_d == 'uint8':
        pass
    elif img_d == 'bool':
        img = img_as_ubyte(img)
        img = invert(img)
        row, col = img.shape
        v = img
        new_img = zeros((img_h, img_w, 3), dtype='uint8')
        new_img[:, :, 0] = v
        new_img[:, :, 1] = v
        new_img[:, :, 2] = v
        img = asarray(new_img, dtype='uint8')
    elif img_d in ('float16', 'float32', 'float64'):
        if amax(img) >= 1 or amin(img) < 0:
            img = clip(img, a_min=0, a_max=1)
        img = img_as_ubyte(img)
    else:
        img = img_as_ubyte(img)
    return img


def _DEV_draw_palette(RGB_pal: List[List[int]], cube_side: int = 2, cube_interval: int = 1,
                bg_col: List[int] = [255, 255, 255], scale_size: int = 50) -> ndarray:

    """
    Palette preview algorithm
    :param RGB_pal:list of lists of colors in RGB uint8 format ([[5,5,5],[6,6,6],[7,7,7]] for example)
    :param cube_side: defines cube`s side
    :param cube_interval: defines interval between cubes and surroundings
    :param bg_col: color of a background of the image
    :param scale_size: a parameter, on which the drawing will be scaled after creating palette
    Output:
    An image of the palette in numpy array form
    """

    img_height = cube_side + cube_interval * 2
    img_width = len(RGB_pal) * cube_side + (len(RGB_pal) + 1) * cube_interval

    bg = image(full((img_height, img_width, 3), bg_col, dtype='uint8'))

    c_cube_MAIN = image(full((cube_side, cube_side, 3), RGB_pal[0], dtype='uint8'))
    c_cube_MAIN.translate((cube_interval, cube_interval))
    c_cube_MAIN_ALPHA = c_cube_MAIN.get_alpha()

    pre_comp = image.merge(bg, c_cube_MAIN, c_cube_MAIN_ALPHA)

    for i in range(1, len(RGB_pal)):
        c_cubeN = image(full((cube_side, cube_side, 3), RGB_pal[i], dtype='uint8'))
        c_cubeN.translate((cube_interval, cube_interval))
        c_cubeN.translate(((cube_interval + cube_side) * i, 0))
        c_cubeN_ALPHA = c_cubeN.get_alpha()
        pre_comp = image.merge(pre_comp, c_cubeN, c_cubeN_ALPHA)

    img = Image.fromarray(image.convert_to_numpy(pre_comp, img_width))
    img = scale(img, scale_size, resample=Image.Resampling.NEAREST)
    font = ImageFont.truetype("arial.ttf", cube_interval * int(scale_size / 2))
    drawer = ImageDraw.Draw(img)
    for i in range(len(RGB_pal)):
        drawer.text(((cube_interval * (i + 1) + cube_side * i) * scale_size, int((cube_interval / 2 - .2) * scale_size)), "Color " + str(i + 1) + ":", font=font, fill='black')

    return array(img)


def _DEV_comp_palette_with_picture(img: ndarray, RGB_palette: List[List[int]]) -> ndarray:
    """
    Composes given image with palette
    Input:
    :param img: a image in numpy array form
    :param RGB_palette: a palette in standard python List[List[int]] format of colors in RGB uint8 format
    Output:
    Will output an image, which will contain given image and palette in numpy array form
    """

    bg_height = round(img.shape[0] * 1.5)
    if img.shape[1] > bg_height:
        bg_height = round(img.shape[1] * 1.5)
    bg_width = bg_height
    bg = full((bg_height, bg_width, 3), 255, dtype='uint8')

    n_img = image(img)
    n_bg = image(bg)

    n_img.translate((round((bg_width - img.shape[1]) / 2), 20))
    n_img_ALPHA = n_img.get_alpha()

    m1 = image.merge(n_bg, n_img, n_img_ALPHA)

    col_interv = 20
    height_interv = img.shape[0] + round(bg_height / 10)
    col_space = (bg_width - col_interv * 2) - (len(RGB_palette) - 1) * col_interv
    if col_space <= 0:
        raise Exception("IMAGE TOO SMALL TO PREVIEW, PLEASE INCREASE THE PIX LIMIT / UPSCALE THE IMAGE")
    cube_side = round(col_space / len(RGB_palette))

    c_cube_MAIN = full((cube_side, cube_side, 3), RGB_palette[0], dtype='uint8')
    n_c_cube_MAIN = image(c_cube_MAIN)
    n_c_cube_MAIN.translate((col_interv, col_interv + height_interv))
    n_c_cube_MAIN_ALPHA = n_c_cube_MAIN.get_alpha()

    pre_comp = image.merge(m1, n_c_cube_MAIN, n_c_cube_MAIN_ALPHA)

    n = 0
    for i in RGB_palette:
        n += 1
        if n == 1:
            continue
        c_cubeN = image(c_cube_MAIN)
        c_cubeN.translate((col_interv, col_interv + height_interv))
        c_cubeN.translate(((col_interv + cube_side) * (n - 1), 0))
        c_cubeN.change_color(i)
        c_cubeN_ALPHA = c_cubeN.get_alpha()
        pre_comp = image.merge(pre_comp, c_cubeN, c_cubeN_ALPHA)

    return image.convert_to_numpy(pre_comp, bg_width)


def output_palette(filepath: str, colors: int = 8, sort_by_lightness: bool = False,
                   pix_limit: int = 200000) -> List[List[int]]:
    """
    Reads an image, reformats it and generates a palette out of it.
    Input:
    :param filepath: a path to the image, which you want to get the palette out of
    :param colors: amount of colors in palette
    :param sort_by_lightness: whether to sort palette by lightness of it`s components or not
    :param pix_limit: a maximum amount of pixels in the formatted image
    Output:
    Will output a standard python list of lists of RGB colors in uint8 format ([[5,5,5],[6,6,6],...] for example)
    """
    img = imread(filepath)

    try:
        img_height, img_width, chan_amount = img.shape
    except ValueError:
        img_height, img_width = img.shape
        chan_amount = 1

    if not (img_width * img_height <= pix_limit and img.dtype == 'uint8' and chan_amount == 3):
        fmtd_img = _DEV_formatting(img, pix_limit, img_width, img_height, chan_amount, img.dtype)
    else:
        fmtd_img = img

    return palette(fmtd_img, colors=colors, SORT_BY_LIGHTNESS=sort_by_lightness)


def output_image_with_palette(filepath: str, colors: int = 8, sort_by_lightness: bool = False,
                   pix_limit: int = 200000) -> ndarray:
    """
    Reads an image, reformats it, generates a palette out of it and then comps it with the image.
    Input:
    :param filepath: a path to the image, which you want to get the palette out of
    :param colors: amount of colors in palette
    :param sort_by_lightness: whether to sort palette by lightness of it`s components or not
    :param pix_limit: a maximum amount of pixels in the formatted image
    Output:
    Will output an image, which will contain given image and it`s palette in numpy array form
    """
    img = imread(filepath)

    try:
        img_height, img_width, chan_amount = img.shape
    except ValueError:
        img_height, img_width = img.shape
        chan_amount = 1

    if not (img_width * img_height <= pix_limit and img.dtype == 'uint8' and chan_amount == 3):
        fmtd_img = _DEV_formatting(img, pix_limit, img_width, img_height, chan_amount, img.dtype)
    else:
        fmtd_img = img

    return _DEV_comp_palette_with_picture(fmtd_img, palette(fmtd_img, colors=colors, SORT_BY_LIGHTNESS=sort_by_lightness))


def output_palette_img(filepath: str, colors: int = 8, sort_by_lightness: bool = False,
                   pix_limit: int = 200000) -> List[List[int]]:
    """
    Reads an image, reformats it, generates a palette out of it and then draws it.
    Input:
    :param filepath: a path to the image, which you want to get the palette out of
    :param colors: amount of colors in palette
    :param sort_by_lightness: whether to sort palette by lightness of it`s components or not
    :param pix_limit: a maximum amount of pixels in the formatted image
    Output:
    Will output an picture, which will contain palette of a given image in numpy array form
    """
    img = imread(filepath)

    try:
        img_height, img_width, chan_amount = img.shape
    except ValueError:
        img_height, img_width = img.shape
        chan_amount = 1

    if not (img_width * img_height <= pix_limit and img.dtype == 'uint8' and chan_amount == 3):
        fmtd_img = _DEV_formatting(img, pix_limit, img_width, img_height, chan_amount, img.dtype)
    else:
        fmtd_img = img

    return _DEV_draw_palette(RGB_pal=palette(fmtd_img, colors=colors, SORT_BY_LIGHTNESS=sort_by_lightness))


"""
MMCQ - Modified Median Cut Quantization
A well known algorithm for quantizing a given image into a set of colors by median of it`s broadest channels
"""
from numpy import ndarray
from typing import List, Tuple, Union, Dict

class VBox(object):

    def __init__(self, r_min: int, r_max: int, g_min: int, g_max: int, b_min: int, b_max: int, histo: Dict[int, int]):
        self.r1 = r_min
        self.r2 = r_max
        self.g1 = g_min
        self.g2 = g_max
        self.b1 = b_min
        self.b2 = b_max
        self.histo = histo

    def average(self) -> List[int]:
        am = 0
        r_sum = 0
        g_sum = 0
        b_sum = 0
        for i in range(self.r1, self.r2):
            for j in range(self.g1, self.g2):
                for k in range(self.b1, self.b2):
                    ijk_key = MAIN.get_color_index(i, j, k)
                    hist_mult = self.histo.get(ijk_key, 0)
                    am += hist_mult
                    r_sum += i * hist_mult * MAIN.MULT
                    g_sum += j * hist_mult * MAIN.MULT
                    b_sum += k * hist_mult * MAIN.MULT
        if am == 0:
            r_avg = int((self.r1 + self.r2) / 2) * MAIN.MULT
            g_avg = int((self.g1 + self.g2) / 2) * MAIN.MULT
            b_avg = int((self.b1 + self.b2) / 2) * MAIN.MULT
        else:
            r_avg = int(r_sum / am)
            g_avg = int(g_sum / am)
            b_avg = int(b_sum / am)
        return [r_avg, g_avg, b_avg]

    def count(self) -> int:
        _sum = 0
        for i in range(self.r1, self.r2):
            for j in range(self.g1, self.g2):
                for k in range(self.b1, self.b2):
                    ijk_key = MAIN.get_color_index(i, j, k)
                    _sum += self.histo.get(ijk_key, 0)
        return _sum

    @property
    def copy(self):
        return VBox(self.r1, self.r2, self.g1, self.g2, self.b1, self.b2, self.histo)

def palette(input_image: ndarray, colors: int=8, SORT_BY_LIGHTNESS: bool=False) -> List[List[int]]:
    """
    Generates palette out of a image (numpy array) (differs from 'output_palette' by not reformating nor decoding the image)
    Input:
    :param input_image: input image(numpy array) to get palette out of
    :param colors: the amount of colors user wants on his palette
    Output:
    Will output a standard python list of lists of RGB colors in uint8 format ([[5,5,5],[6,6,6],...] for example)
    """
    img = []
    for row in input_image.tolist():
        for pix in row:
            if not pix == [0, 0, 0] or pix[0] + pix[1] + pix[2] >= 750:
                img.append(pix)

    histo = MAIN.get_histo(img)

    if len(histo) <= colors:
        print('IMAGE DOES NOT CONTAIN ENOUGH INFO FOR PROPER PALETTE')
        vbox_mass = []
        for i in range(colors):
            vbox_mass.append(input_image.tolist()[MAIN.get_random_indx(len(input_image), i+1)]
                             [MAIN.get_random_indx(len(input_image), int(i+1)*2)])
        return vbox_mass


    vbox = MAIN.get_vbox(img, histo)

    def iter_(VirtualBox: VBox, colors: int) -> Tuple[List[VBox], int]:
        vbox_list = [VirtualBox]
        n = 1
        i = 0
        input_highres = True
        while n < colors:
            undefined_colors = 0
            vbox1, vbox2, indx, out_highres = MAIN.get_median_cut(vbox_list, i, input_highres)
            if vbox2 == 0:
                del vbox_list[indx]
                vbox_list.append(vbox1)
                i += 1
            elif vbox1 == 1 and vbox2 == 1 and indx == 1:
                if out_highres:
                    input_highres = False
                    continue
                else:
                    undefined_colors = colors-n
                    break
            else:
                del vbox_list[indx]
                vbox_list.append(vbox1)
                vbox_list.append(vbox2)
                n += 1
                i = 0
        return vbox_list, undefined_colors

    vbox_mass, undefined_colors = iter_(vbox, colors)
    vbox_mass.sort(key=lambda x: x.count(), reverse=True)
    vbox_mass = list(map(lambda x: x.average(), vbox_mass))
    for i in range(undefined_colors):
        vbox_mass.append(img[MAIN.get_random_indx(len(img), i+1, random_number)])
    if SORT_BY_LIGHTNESS:
        vbox_mass.sort(key=lambda x: x[0] + x[1] + x[2], reverse=True)
    return vbox_mass


class MAIN(object):
    MULT = 8
    COL_ACCURACY = 2
    MAX_ITERATION = 100

    @staticmethod
    def get_median_cut(vbox_list: List[VBox], i:int, highres:bool) -> Tuple[Union[VBox, int], Union[VBox, int], int, bool]:
        vbox_indx = 0
        MRO = [0, 0, 0]
        for vbox in vbox_list:
            rRange = vbox.r2 - vbox.r1
            gRange = vbox.g2 - vbox.g1
            bRange = vbox.b2 - vbox.b1
            if rRange >= gRange and rRange >= bRange:
                cut_chan = 'r'
                mRange = rRange
            elif gRange >= rRange and gRange >= bRange:
                cut_chan = 'g'
                mRange = gRange
            else:
                cut_chan = 'b'
                mRange = bRange
            if highres:
                if MRO[0] < mRange:
                    MRO = [mRange, cut_chan, vbox_indx]
            else:
                if MRO[0] < vbox.count():
                    MRO = [vbox.count(), cut_chan, vbox_indx]
            vbox_indx += 1
        cut_chan = MRO[1]
        vbox_indx = MRO[2]
        mass = range(getattr(vbox_list[vbox_indx], cut_chan + '1'),
                     getattr(vbox_list[vbox_indx], cut_chan + '2'))
        l = len(mass)
        if l % 2:
            c_cut = mass[int(l / 2)]
        else:
            c_cut = round((mass[int(l / 2) - 1] + mass[int(l / 2)]) / 2)
        vbox1 = vbox_list[vbox_indx].copy
        vbox2 = vbox_list[vbox_indx].copy
        setattr(vbox1, cut_chan + '2', c_cut)
        setattr(vbox2, cut_chan + '1', c_cut + 1)

        if i <= MAIN.MAX_ITERATION:
            if vbox2.count() <= MAIN.COL_ACCURACY:
                if vbox1.count() <= MAIN.COL_ACCURACY:
                    print("IMAGE TOO SMALL/TOO MONOTONIC/TOO RANDOM TO PROPERLY EXPORT THE DESIRED AMOUNT OF COLORS")
                    if vbox1.count() == 1 and vbox2.count() == 1:
                        return vbox1, vbox2, vbox_indx, True
                    else:
                        if highres:
                            return 1,1,1,True
                        else:
                            return 1,1,1,False
                else:
                    return vbox1, 0, vbox_indx, True
            else:
                if vbox1.count() <= MAIN.COL_ACCURACY:
                    return vbox2, 0, vbox_indx, True
                else:
                    return vbox1, vbox2, vbox_indx, True
        else:
            return vbox1, vbox2, vbox_indx, True

    @staticmethod
    def get_color_index(r: int, g: int, b: int) -> int: #basically convert 3 8-bit values into 1 24-bit one
        return (r << MAIN.MULT) + (g << (MAIN.MULT * 2)) + b

    @staticmethod
    def get_histo(img: List[List[int]]) -> Dict[int, int]:
        histo = dict()
        for pix in img:
            r = int(pix[0] / MAIN.MULT)
            g = int(pix[1] / MAIN.MULT)
            b = int(pix[2] / MAIN.MULT)
            histo_key = MAIN.get_color_index(r, g, b)
            histo[histo_key] = histo.setdefault(histo_key, 0) + 1
        return histo

    @staticmethod
    def get_vbox(img: List[List[int]], histo: Dict[int, int]) -> VBox:
        r_min = 255
        r_max = 0
        g_min = 255
        g_max = 0
        b_min = 255
        b_max = 0
        for pix in img:
            r = int(pix[0] / MAIN.MULT)
            g = int(pix[1] / MAIN.MULT)
            b = int(pix[2] / MAIN.MULT)
            r_min = min(r_min, r)
            r_max = max(r_max, r)
            g_min = min(g_min, g)
            g_max = max(g_max, g)
            b_min = min(b_min, b)
            b_max = max(b_max, b)
        return VBox(r_min, r_max, g_min, g_max, b_min, b_max, histo)

    @staticmethod
    def get_random_indx(img_size: int, i: int) -> int: #not really working algorithm, for the proper use
        pi = '31459265358979323846'
        seed = pi[i] * str(str(img_size)[0]*i)[0] // 81
        return int(img_size * seed) - 1



from setuptools import setup, find_packages
import codecs
import os

setup(
    name = "colorpalette-master",
    version = "0.0.4",
    author = "Victor Andrievskiy",
    author_email = "superavb222@gmail.com",
    description= "Scripts for generating a color palette out of an image",
    long_description= "Scripts for generating a color palette out of an image, based on median cut algorithm. Used in my college project PixPalette.",
    packages=find_packages(),
    install_requires = ['scikit-image', 'numpy', 'Pillow==9.4.0'],
    url = "https://github.com/gejirz/colorpalette-master",
    keywords=['python', 'color palette', 'median cut quantization', 'PixPalette', 'median cut'],
    license= "MIT License"
)


from numpy import ndarray, asarray
from typing import Tuple, List, Dict


class image(object):
    """
    A unoptimized class used to create images.
    Basically you first provide the information for the image in numpy array form, then it is converted into
    Dict[Tuple[int, int], List[int]] form, where Tuple defines the x, y coordinates of a pixel, and List is the
    pixel in RGB uint8 form ({(0,0):[255,255,255],...} for example)
    """

    def __init__(self, image: ndarray):
        self.img = dict()
        x = -1
        y = -1
        for row in image.tolist():
            y += 1
            for pix in row:
                x += 1
                if x >= image.shape[1]:
                    x = 0
                self.img[x,y] = pix


    def translate(self, coord: Tuple[int, int]=(0,0)):
        new_img = dict()
        for k,v in self.img.items():
            x,y = k
            new_x = x
            new_y = y
            new_x += int(coord[0])
            new_y += int(coord[1])
            new_img[new_x,new_y] = v
        self.img = new_img


    def items(self):
        return self.img.items()

    def pop(self, index: int, out: int=0) -> int:
        return self.img.pop(index, out)

    def change_color(self, color: List[int]=[0,0,0]) -> Dict[Tuple[int, int], List[int]]:
        for k,v in self.img.items():
            self.img[k] = color
        return self.img

    def get_alpha(self) -> Dict[Tuple[int, int], List[int]]:
        alpha = dict()
        for k,v in self.img:
            if v == 0 or v == [0,0,0]:
                alpha[k] = 0
            else:
                alpha[k] = 1
        return alpha


    @staticmethod
    def merge(img_B: Dict[Tuple[int, int], List[int]], img_A: Dict[Tuple[int, int], List[int]], alpha: Dict[Tuple[int, int], List[int]] = dict()) -> Dict[Tuple[int, int], List[int]]:
        comp = dict()
        for k_B,v_B in img_B.items():
            a = alpha.pop((k_B), 0)
            if type(a) == list:
                for x in a:
                    a = x
            r_B, g_B, b_B = v_B
            r_A, g_A, b_A = img_A.pop(k_B, [0,0,0])
            comp[k_B] = [r_A + r_B * (1 - a), g_A + g_B * (1 - a), b_A + b_B * (1 - a)]
        return comp


    @staticmethod
    def convert_to_numpy(img: Dict[Tuple[int, int], List[int]], width: int) -> ndarray:
        l = list()
        l2 = list()
        n_cycleAm = 0
        for k,v in img.items():
            l2.append(v)
            n_cycleAm += 1
            if n_cycleAm >= width:
                l.append(l2[:])
                n_cycleAm = 0
                l2.clear()
        out_img = asarray(l, dtype='uint8')
        return out_img